# AUTOGENERATED! DO NOT EDIT! File to edit: 03_pytorch_discrete_choice.ipynb (unless otherwise specified).

__all__ = ['log_loss']

# Cell
# Py3 Built-in module for type-hint annotations
from typing import Dict, Tuple, List, Callable, Optional
# Third-party modules for:
# class definition
import attr
# Optimization
import scipy.optimize as opt
# Numerical computation
import torch
import torch.nn as nn
import numpy as np
# Utility functions for pytorch optimization
import botorch.optim.numpy_converter as numpy_converter
import botorch.optim.utils as optim_utils
from botorch.optim.numpy_converter import TorchAttr
# Input/Output of data
import pandas as pd

# Cell
def log_loss(probs: torch.Tensor,
             targets: torch.Tensor) -> torch.Tensor:
    """
    Computes the log-loss (i.e., the negative log-likelihood) for given
    long-format tensors of probabilities and choice indicators.

    Parameters
    ----------
    probs : 1D torch.Tensor.
        The probability of choosing each row's alternative for the given choice situation.
    targets : 1D torch.Tensor.
        A Tensor of zeros and ones indicating the chosen row for each choice situation. Should have the same size as `probs`.

    Returns
    -------
    neg_log_likelihood : scalar torch.Tensor.
        The negative log-likelihood computed from `probs` and `targets`.
    """
    log_likelihood = torch.sum(targets * torch.log(probs))
    return -1 * log_likelihood


# AUTOGENERATED! DO NOT EDIT! File to edit: 02_MNL_swissmetro.ipynb (unless otherwise specified).

__all__ = []

# Cell
import pickle
import numpy as np
import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
import biogeme.models as models
from biogeme.expressions import Beta

# Cell
#load package...


# AUTOGENERATED! DO NOT EDIT! File to edit: 01_statsmodels_intro.ipynb (unless otherwise specified).

__all__ = []

# Cell
import numpy as np
import pandas as pd
import patsy
from scipy import stats
import statsmodels.api as sm
from statsmodels.formula.api import logit
import matplotlib.pyplot as plt


# AUTOGENERATED! DO NOT EDIT! File to edit: 01_statsmodels_MNL.ipynb (unless otherwise specified).

__all__ = []

# Cell
import numpy as np
import statsmodels.api as sm
import pandas as pd
import patsy


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"MNL": "03_pytorch_MNL.ipynb",
         "load_model": "03_pytorch_MNL.ipynb",
         "build_model": "03_pytorch_MNL.ipynb",
         "log_loss": "03_pytorch_discrete_choice.ipynb",
         "get_act": "10_tastenet_mnl.ipynb",
         "ChoiceFlex": "10_tastenet_mnl.ipynb",
         "Utility": "10_tastenet_mnl.ipynb",
         "TasteParams": "10_tastenet_mnl.ipynb",
         "ChoiceDataset": "10_tastenet_mnl.ipynb",
         "getAttribute": "10_tastenet_mnl.ipynb",
         "selectZ": "10_tastenet_mnl.ipynb"}

modules = ["statsmodels_MNL.py",
           "statsmodels_intro.py",
           "MNL_swissmetro.py",
           "pytorch_MNL.py",
           "pytorch_discrete_choice.py",
           "tastenet_mnl.py"]

doc_url = "https://danhphan.github.io/choicenet/"

git_url = "https://github.com/danhphan/choicenet/tree/master/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: 03_pytorch_MNL.ipynb (unless otherwise specified).

__all__ = ['MNL', 'load_model', 'build_model']

# Cell
import torch

import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torch.autograd import Variable

import math
# serialization and deserialization of model
import pickle

import numpy as np
import pandas as pd

class MNL(nn.Module):
    '''
        The Multinomial Logistic Regression model implemented with Pytorch
    '''
    def __init__(self, feature_list):
        super(MNL, self).__init__()

        self.feature_list = feature_list
        input_dim = len(feature_list)
        # a linear layer without bias
        self.linear = torch.nn.Linear(input_dim, 1, bias=False)
        self.softmax = torch.nn.Softmax()


    def forward(self, x):
        # expect the input to be a session of alternatives
        util_values = self.linear(x)

        #!! a trik to prevent the underflow (divide by zero) in the softmax later
        util_values = util_values + 2

        # transpose the result vector before the softmax
        util_values = torch.t(util_values)

        # convert the softmax values to binary values
        #max_values, indices = self.softmax(util_values).max()

        #results = np.zeros(len(x))
        #results[indices] = 1
        #results = np.transpose(results)

        return torch.t(self.softmax(util_values))


    def l1_loss(self, l1_weight=0.01):
        '''
            Return: L1 regularization on all the parameters
        '''
        params_list = []
        for param in self.parameters():
            params_list.append(param.view(-1))
        torch_params = torch.cat(params_list)

        return l1_weight * (torch.abs(torch_params).sum())


    def l2_loss(self, l2_weight=0.01):
        '''
            Return: L2 regularization on all the parameters
        '''
        params_list = []
        for param in self.parameters():
            params_list.append(param.view(-1))
        torch_params = torch.cat(params_list)

        return l2_weight * (torch.sqrt(torch.pow(torch_params, 2).sum()))


    def train(self, loss, optimizer, x_val, y_val,
              l1_loss_weight = 0,  # when zero, no L1 regularization
              l2_loss_weight = 0,
              gpu=False):
        """
            Train the model with a batch (in our case, also a session) of data
        """
        # expect y_val to be of one_dimension
        y_val = y_val.reshape(len(y_val), 1)

        tensorX = torch.from_numpy(x_val).double()
        tensorY = torch.from_numpy(y_val).double()

        if (gpu):
            dtype = torch.cuda.DoubleTensor
        else:
            dtype = torch.DoubleTensor

        # input variable
        x = Variable(tensorX.type(dtype), requires_grad=False)
        # target variable
        y = Variable(tensorY.type(dtype), requires_grad=False)

        # Forward to calculate the losses
        fx = self.forward(x)
        data_loss = loss.forward(fx, y)

        # optional: add L1 or L2 penalities for regularization
        if (l1_loss_weight != 0):
            l1_loss = self.l1_loss(l1_loss_weight)
            output = data_loss + l1_loss

        elif (l2_loss_weight != 0):
            l2_loss = self.l2_loss(l2_loss_weight)
            output = data_loss + l2_loss

        else:
            output = data_loss

        # Underflow in the loss calculation
        if math.isnan(output.data[0]):
            raise ValueError('NaN detected')
            #return output.data[0]

        if (isinstance(optimizer, torch.optim.LBFGS)):
            def closure():
                optimizer.zero_grad()
                output.backward(retain_graph=True)
                return output

            optimizer.step(closure)
        else:
            # Reset gradient
            optimizer.zero_grad()
            # Backward
            output.backward()
            # Update parameters
            optimizer.step()

        # return the cost
        return output.data[0]


    def predict(self, x_val, binary=False):
        '''
            Give prediction for alternatives within a single session
            x_val: DataFrame, or np.ndarray
            return: numpy
        '''
        is_gpu = self.get_params()[0].is_cuda

        if isinstance(x_val, pd.DataFrame):
            tensorX = torch.from_numpy(x_val.values).double()
        else:
            tensorX = torch.from_numpy(x_val).double()

        if (is_gpu):
            x = Variable(tensorX.type(torch.cuda.DoubleTensor), requires_grad=False)
        else:
            x = Variable(tensorX, requires_grad=False)

        output = self.forward(x)

        if (is_gpu):
            # get the data from the memory of GPU into CPU
            prob = output.cpu().data.numpy()
        else:
            prob = output.data.numpy()

        if (binary):
            # convert the softmax values to binary values
            max_indice = prob.argmax(axis=0)
            ret = np.zeros(len(prob))
            ret[max_indice] = 1
            return ret
        else:
            return prob


    def get_params(self):
        '''
            Return the Variable of the MNL parameters,
              which can be updated manually.
        '''
        for name, param in self.named_parameters():
            if param.requires_grad and name == 'linear.weight':
                return param
        return None


    def print_params(self):
        '''
            Print all the parameters within the model
        '''
        params = self.get_params()[0]

        if (params.is_cuda):
            values = params.cpu().data.numpy()
        else:
            values = params.data.numpy()

        for index, feature in enumerate(self.feature_list):
            print(feature, ':', values[index])


    def get_feature_weight(self, feature_name):
        ''' Retrieve the weight of the desired feature '''
        params = self.get_params()[0]

        if (params.is_cuda):
            param_values = params.cpu().data.numpy()
        else:
            param_values = params.data.numpy()

        for index, feature in enumerate(self.feature_list):
            if (feature_name == feature):
                return param_values[index]

        # did not find the specified feature
        return None


    def get_feature_weights(self):
        ''' get the dictionary of feature weights '''
        params = self.get_params()[0]

        if (params.is_cuda):
            param_values = params.cpu().data.numpy()
        else:
            param_values = params.data.numpy()

        feature_weights = {}

        for index, feature in enumerate(self.feature_list):
            feature_weights[feature] = param_values[index]

        return feature_weights


    def set_feature_weight(self, feature_name, value):
        ''' Reset the specified feature weight
        '''
        params = self.get_params()[0]
        is_found = False

        try:
            for index, feature in enumerate(self.feature_list):
                if (feature_name == feature):
                    is_found = True
                    # override the parameters within the model
                    params[index] = value
        except RuntimeError as e:
            #print('RuntimeError: ', e)
            #print('One can ignore this error, since the parameters are still updated !')
            pass

        return is_found


    def save(self, file_name):
        '''
            Serialize the model object into a file
        '''
        with open(file_name, mode='wb') as model_file:
            pickle.dump(self, model_file)
            print('save model to ', file_name)


    def set_train_config(self, train_config):
        '''
            Set the training configs along with the model,
             so that it can be serialized together with the model.
        '''
        self.train_config = train_config


    def get_train_config(self):
        return self.train_config


def load_model(pickle_file):
    """
        Instantialize a model from its pickle file.
    """
    with open(pickle_file, 'rb') as inp:
        try:
            # python 3
            model = pickle.load(inp, encoding='bytes')
        except:
            # python 2
            model = pickle.load(inp)

        print('load model from ', pickle_file)
        return model


def build_model(input_dim):
    '''
        Another way to build the model.
    '''
    model = torch.nn.Sequential()
    model.add_module("linear",
                     torch.nn.Linear(input_dim, 1, bias=False))

    # We need the softmax layer here for the binary cross entropy later
    model.add_module('softmax', torch.nn.Softmax())

    return model





# AUTOGENERATED! DO NOT EDIT! File to edit: 10_tastenet_mnl.ipynb (unless otherwise specified).

__all__ = ['get_act', 'ChoiceFlex', 'Utility', 'TasteParams', 'ChoiceDataset', 'getAttribute', 'selectZ']

# Cell

import torch
import torch.nn as nn
from collections import OrderedDict
import torch.nn.functional as F

def get_act(nl_func):
    if nl_func=="tanh":
        return nn.Tanh()
    elif nl_func == "relu":
        return nn.ReLU()
    elif nl_func == "sigmoid":
        return nn.Sigmoid()
    else:
        return None

class ChoiceFlex(nn.Module):
    def __init__(self, args):
        super(ChoiceFlex, self).__init__()
        self.params_module = TasteParams(args.layer_sizes, args)
        self.util_module = Utility(args)
        self.args = args

    def forward(self, z, x, av):
        b = self.params_module(z) # taste parameters, (N,8)
        b = self.constraints(b)  ## this is another way to include constraint:  using transformation to include constraints
        v = self.util_module(x,b) #no softmax here
        exp_v = torch.exp(v)
        exp_v_av = exp_v * av

        prob = exp_v_av/exp_v_av.sum(dim=1).view(-1,1) # prob (N,J)

        return prob, None

    def constraints(self,b):
        '''
            Put transformation for the sake of constraints on the value of times
        '''
        if self.args.transform=='relu':
            return torch.cat([-F.relu(-b[:,:-3]),b[:,-3:]],dim=1)
        elif self.args.transform == 'exp':
            return torch.cat([-torch.exp(-self.args.mu * b[:,:-3]),b[:,-3:]],dim=1) # the last 3 dim of b are under constraints
        else:
            return b

    def getParameters(self):
        '''
        get coef and bias of the TasteParams of the model
        '''
        count = 0
        bias = []
        coef = []
        for params in self.parameters():
            if count % 2==0:
                coef.append(params)
            else:
                bias.append(params)
            count += 1
        return coef, bias

    def L2Norm(self):
        '''
        L2 norm, not including bias
        '''
        coef, bias = self.getParameters()
        norm = torch.zeros(1)
        for params in coef:
            norm += (params**2).sum()
        return norm

    def L1Norm(self):
        '''
        L1 norm, not including bias
        '''
        coef, bias = self.getParameters()
        norm = torch.zeros(1)
        for params in coef:
            norm += (torch.abs(params).sum())
        return norm

class Utility(nn.Module):
    def __init__(self, args):
        super(Utility, self).__init__()
        self.args = args
        self.index = OrderedDict(zip(['TRAIN_TT', 'SM_TT', 'CAR_TT', 'TRAIN_HE', 'SM_HE', 'SM_SEATS', 'TRAIN_ASC', 'SM_ASC'], range(8)))


    def forward(self, x, b):
        '''
        x: attributes of each alternative,
           including the intercept (N,K+1,J)  J alternatives, each have K+1 attributes including 1 for intercept.
        b: taste parameters (K+1,J)  Some paramters are constant, some come from neural network hidden layer.
        '''
        index = self.index
        N = len(b)
        # last hidden nodes correspond to b_names
        v = torch.zeros(N,3)
        v[:,0] = torch.ones(N) * b[:,index["TRAIN_ASC"]] + x["TRAIN"]["TRAIN_TT"]*b[:,index["TRAIN_TT"]] + x["TRAIN"]["TRAIN_HE"]*b[:,index["TRAIN_HE"]] - x["TRAIN"]["TRAIN_CO"]
        v[:,1] = torch.ones(N) * b[:,index["SM_ASC"]] + x["SM"]["SM_TT"]*b[:,index["SM_TT"]] + x["SM"]["SM_HE"]*b[:,index["SM_HE"]] + x["SM"]["SM_SEATS"]*b[:,index["SM_SEATS"]] - x["SM"]["SM_CO"]
        v[:,2] = x["CAR"]["CAR_TT"]*b[:,index["CAR_TT"]] - x["CAR"]["CAR_CO"]

        return v

class TasteParams(nn.Module):
    '''
    Network for tastes
    Structure: MLP
    '''
    def __init__(self, layer_sizes, args):
        super(TasteParams, self).__init__()
        self.seq = nn.Sequential()
        for i, (in_size, out_size) in enumerate(zip(layer_sizes[:-1], layer_sizes[1:])): #layer_sizes = [input_size,..., K(number of parameters)]
            self.seq.add_module(name="L%i"%(i+1), module=nn.Linear(in_size, out_size, bias=True))
            if i<len(layer_sizes)-2:
                self.seq.add_module(name="A%i"%(i+1), module=get_act(args.act_func))
        self.args = args

    def forward(self,z):
        '''
        Parameters:
            z: (N,D) # batch size, input dimension
        Returns:
            V: (N,K) # taste parameters
        '''
        N,D = z.size()
        return self.seq(z) # (N,K)

# Cell
import pickle
from torch.utils.data import Dataset
import torch
import numpy as np

class ChoiceDataset(Dataset):
    """Choice dataset"""

    def __init__(self, data_path, data_file):
        """
        Parameters:
            data_file (string): name of data pickle file
        """
        data = pickle.load(open(data_path + "/" + data_file, "rb"))

        self.x = torch.Tensor(data["x"])
        self.x_names = data["x_names"]
        self.N = len(self.x)

        dic_attr = \
        {"TRAIN": ["TRAIN_TT", "TRAIN_HE", "TRAIN_CO"], \
         "SM": ["SM_TT", "SM_HE", "SM_CO", "SM_SEATS"],\
         "CAR": ["CAR_TT", "CAR_CO"]}

        self.x_dict = {}
        for mode in dic_attr:
            self.x_dict[mode] = {}
            for attr in dic_attr[mode]:
                self.x_dict[mode].update({attr: getAttribute(self.x, self.x_names, attr)})

        self.y = torch.LongTensor(data["y"])-1 # N

        # Availability
        self.av = torch.cat([torch.ones(self.N,2),
                             torch.Tensor(data["car_av"]).view(self.N,1)], dim=1) # (N,3) av for all modes

        # all z
        self.z_all_names = data['z_names']
        self.z_levels = data['z_levels']
        self.z_all = torch.Tensor(data['z']) # N,D socio-demo variables

        # select z
        self.z_names = ["MALE_1", "AGE_1", "AGE_2", "AGE_3", "AGE_4", \
               "INCOME_1", "INCOME_2", "INCOME_3", "FIRST_1", "WHO_1", "WHO_2", \
               "PURPOSE_1", "PURPOSE_2", "PURPOSE_3", "LUGGAGE_1", "LUGGAGE_2", "GA_1"]
        self.z = selectZ(self.z_all, self.z_names, self.z_all_names)

        _, self.D = self.z.size() # z size = (N,D)


    def __len__(self):
        return self.N

    def __getitem__(self, idx):
        '''
        Get the sample given its idx in the list
        '''
        x = {} # x is a dictionary!
        for mode in self.x_dict:
            x[mode] = {}
            for name in self.x_dict[mode]:
                x[mode][name] = self.x_dict[mode][name][idx]
        return {"x": x, "y": self.y[idx], "z":self.z[idx], "av": self.av[idx]}

def getAttribute(x, x_names, name):
    return x[:,x_names.index(name)]

def selectZ(z,z_selected, z_names):
    ind = []
    for var in z_selected:
        ind.append(z_names.index(var))
    return z[:,np.array(ind).astype(int)]
